//	Metaballs II
//	www.paulsprojects.net
//
//	Copyright (c) 2008, Paul Baker
//	All rights reserved.
//
//	Redistribution and use in source and binary forms, with or without modification,
//	are permitted provided that the following conditions are met:
//
//	    * Redistributions of source code must retain the above copyright notice,
//	      this list of conditions and the following disclaimer.
//	    * Redistributions in binary form must reproduce the above copyright notice,
//	      this list of conditions and the following disclaimer in the documentation
//	      and/or other materials provided with the distribution.
//	    * Neither the name of paulsprojects.net nor the names of its contributors
//	      may be used to endorse or promote products derived from this software
//	      without specific prior written permission.
//
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//	DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//	ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;

using SlimDX;
using SlimDX.Direct3D;
using SlimDX.Direct3D9;

namespace Metaballs_II
{
    unsafe class CubeGrid : IDisposable
	{
		private const float minX = -10.0f, maxX = 10.0f;
		private const int numCubesX = 79, numCubesY = 79, numCubesZ = 39;
		private const float minY = (minX * numCubesY) / numCubesX, maxY = (maxX * numCubesY) / numCubesX;
		private const float minZ = (minX * numCubesZ) / numCubesX, maxZ = (maxX * numCubesZ) / numCubesX;

		private const int numVerticesX = numCubesX + 1;
		private const float stepX = (maxX - minX) / (numVerticesX - 1);
				
		private const int numVerticesY = numCubesY + 1;
		private const float stepY = (maxY - minY) / (numVerticesY - 1);

		private const int numVerticesZ = numCubesZ + 1;
		private const float stepZ = (maxZ - minZ) / (numVerticesZ - 1);

		private const int numVertices = numVerticesX * numVerticesY * numVerticesZ;
		private const int numCubes = numCubesX * numCubesY * numCubesZ;

		private static readonly float * vertexXs = AllocateAlignedFloats(numVertices);
		private static readonly float * vertexYs = AllocateAlignedFloats(numVertices);
		private static readonly float * vertexZs = AllocateAlignedFloats(numVertices);

		private static readonly int[,] cubeVertexIndices = new int[numCubes, 8];
		
		private readonly float * vertexValues = AllocateAlignedFloats(numVertices);

		private readonly float * vertexNormalXs = AllocateAlignedFloats(numVertices);
		private readonly float * vertexNormalYs = AllocateAlignedFloats(numVertices);
		private readonly float * vertexNormalZs = AllocateAlignedFloats(numVertices);

		[StructLayout(LayoutKind.Sequential)]
		private struct SurfaceVertex
		{
			public Vector3 position;
			public Vector3 normal;

			public static VertexFormat Format
			{
				get
				{
					return VertexFormat.Position | VertexFormat.Normal;
				}
			}
		}

		private const ushort maxSurfaceVertices = 10000;
		private SurfaceVertex[] surfaceVertices = new SurfaceVertex[maxSurfaceVertices];
		private ushort surfaceVertexIndex;
		private VertexBuffer surfaceVertexBuffer;

		private const int maxSurfaceIndices = 50000;
		private ushort[] surfaceIndices = new ushort[maxSurfaceIndices];
		private int surfaceIndexIndex;
		private IndexBuffer surfaceIndexBuffer;

		/// <summary>
		/// Allocate a page-aligned array of single-precision floating point numbers
		/// </summary>
		/// <param name="count">The size of the array</param>
		/// <returns>A pointer to the allocated array</returns>
		private static float * AllocateAlignedFloats(int count)
		{
			IntPtr ptr = Native.VirtualAlloc((IntPtr)0, (UIntPtr)(count * sizeof(float)), Native.MEM_COMMIT, Native.PAGE_READWRITE);

			if(ptr == (IntPtr)0)
			{
				throw new System.OutOfMemoryException();
			}

			return (float *)ptr;
		}

		/// <summary>
		/// Free a page-aligned array of single-precision floating point numbers
		/// </summary>
		/// <param name="ptr">A pointer to the array</param>
		/// <param name="count">The size of the array</param>
		private static void FreeAlignedFloats(float * ptr, int count)
		{
			Native.VirtualFree((IntPtr)ptr, (UIntPtr)(count * sizeof(float)), Native.MEM_RELEASE);
		}

		static CubeGrid()
		{
			int vertex = 0;
			for(int i = 0; i < numVerticesX; ++i)
			{
				for(int j = 0; j < numVerticesY; ++j)
				{
					for(int k = 0; k < numVerticesZ; ++k)
					{
						vertexXs[vertex] = minX + i * stepX;
						vertexYs[vertex] = minY + j * stepY;
						vertexZs[vertex] = minZ + k * stepZ;

						++vertex;
					}
				}
			}

			int cube = 0;
			for(int i = 0; i < numCubesX; ++i)
			{
				for(int j = 0; j < numCubesY; ++j)
				{
					for(int k = 0; k < numCubesZ; ++k)
					{
						cubeVertexIndices[cube, 0] = (i * numVerticesY + j) * numVerticesZ + k;
						cubeVertexIndices[cube, 1] = (i * numVerticesY + j) * numVerticesZ + k + 1;
						cubeVertexIndices[cube, 2] = (i * numVerticesY + j + 1) * numVerticesZ + k + 1;
						cubeVertexIndices[cube, 3] = (i * numVerticesY + j + 1) * numVerticesZ + k;
						cubeVertexIndices[cube, 4] = ((i + 1) * numVerticesY + j) * numVerticesZ + k;
						cubeVertexIndices[cube, 5] = ((i + 1) * numVerticesY + j) * numVerticesZ + k + 1;
						cubeVertexIndices[cube, 6] = ((i + 1) * numVerticesY + j + 1) * numVerticesZ + k + 1;
						cubeVertexIndices[cube, 7] = ((i + 1) * numVerticesY + j + 1) * numVerticesZ + k;
						
						++cube;
					}
				}
			}
		}

		private bool disposed = false;

		~CubeGrid()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if(!disposed)
			{
				if(disposing)
				{
					if(surfaceVertexBuffer != null)	surfaceVertexBuffer.Dispose();
					if(surfaceIndexBuffer != null)	surfaceIndexBuffer.Dispose();
				}

				FreeAlignedFloats(vertexValues, numVertices);
				FreeAlignedFloats(vertexNormalXs, numVertices);
				FreeAlignedFloats(vertexNormalYs, numVertices);
				FreeAlignedFloats(vertexNormalZs, numVertices);

				disposed = true;
			}
		}

		public class CalculateVertexValuesStartData
		{
			public CalculateVertexValuesStartData(Metaball[] metaballs_, AABoundingBox boundingBox_, ManualResetEvent doneEvent_)
			{
				metaballs = metaballs_;
				boundingBox = boundingBox_;
				doneEvent = doneEvent_;
			}

			public Metaball[] metaballs;
			public AABoundingBox boundingBox;
			public ManualResetEvent doneEvent;
		}
		
		/// <summary>
		/// Calculate the values and normals at each vertex of the grid
		/// </summary>
		/// <param name="obj">An object of type CalculateVertexValuesStartData</param>
		public void CalculateVertexValues(object obj)
		{
#if PROFILE
			Program.profiler.BeginSection("CalculateVertexValues");
#endif
			CalculateVertexValuesStartData startData = (CalculateVertexValuesStartData)obj;
			Metaball[] metaballs = startData.metaballs;
			AABoundingBox boundingBox = startData.boundingBox;
			ManualResetEvent doneEvent = startData.doneEvent;

			int minCubeX, minCubeY, minCubeZ, maxCubeX, maxCubeY, maxCubeZ;
			PointToCubeXYZ(boundingBox.min, out minCubeX, out minCubeY, out minCubeZ);
			PointToCubeXYZ(boundingBox.max, out maxCubeX, out maxCubeY, out maxCubeZ);

			fixed(Metaball * pMetaballs = metaballs)
			{
				MetaCalc.CalculateVertexValues(	numVerticesX, numVerticesY, numVerticesZ,
												vertexXs, vertexYs, vertexZs,
												vertexValues, vertexNormalXs, vertexNormalYs, vertexNormalZs,
												metaballs.Length, pMetaballs, minCubeX, minCubeY, maxCubeX, maxCubeY);
			}
#if PROFILE
			Program.profiler.EndSection("CalculateVertexValues");
#endif
			doneEvent.Set();
		}

		public class CalculateSurfaceVerticesStartData
		{
			public CalculateSurfaceVerticesStartData(	Metaball[] metaballs_, float threshold_, ManualResetEvent doneEvent_)
			{
				metaballs = metaballs_;
				threshold = threshold_;
				doneEvent = doneEvent_;
			}

			public Metaball[] metaballs;
			public float threshold;
			public ManualResetEvent doneEvent;
		}

		private int frame = 0;
		private int[] cubesCalculated = new int[numCubes];
		private List<int> cubesToCalculate = new List<int>();
		
		/// <summary>
		/// Calculate the positions and normals for all of the surface vertices
		/// </summary>
		/// <param name="obj">An object of type CalculateSurfaceVerticesStartData</param>
		public void CalculateSurfaceVertices(object obj)
		{
#if PROFILE
			Program.profiler.BeginSection("CalculateSurfaceVertices");
#endif
			CalculateSurfaceVerticesStartData startData = (CalculateSurfaceVerticesStartData)obj;
			Metaball[] metaballs = startData.metaballs;
			float threshold = startData.threshold;
			ManualResetEvent doneEvent = startData.doneEvent;

			//Metaballs' radii == 0 => First frame => No surface vertex values have been calculated
			if(metaballs[0].radiusSquared != 0.0f)
			{
				++frame;
				surfaceVertexIndex = 0;
				surfaceIndexIndex = 0;

				foreach(Metaball metaball in metaballs)
				{
					int centreCubeX, centreCubeY, centreCubeZ;
					PointToCubeXYZ(metaball.position, out centreCubeX, out centreCubeY, out centreCubeZ);

					int centreCube = (centreCubeX * numCubesY + centreCubeY) * numCubesZ + centreCubeZ;

					int cube = centreCube;
					bool surfaceFound = false;
					byte verticesInsideSurfaceCube = 0x00;
					for(;;)
					{
						if(cubesCalculated[cube] == frame)
							break;

						cubesCalculated[cube] = frame;
						byte verticesInside;
						if(AddSurfaceVertices(cube, threshold, out verticesInside))
						{
							surfaceFound = true;
							verticesInsideSurfaceCube = verticesInside;
							break;
						}

						cube -= numCubesY * numCubesZ;
						Debug.Assert(cube >= 0);
					}

					if(surfaceFound)
					{
						cubesToCalculate.Clear();
						AddNeighbouringCubesToList(cubesToCalculate, cube, verticesInsideSurfaceCube, cubesCalculated);

						for(int cubeIndex = 0; cubeIndex != cubesToCalculate.Count; ++cubeIndex)
						{
							cube = cubesToCalculate[cubeIndex];

							if(cubesCalculated[cube] == frame)
								continue;

							cubesCalculated[cube] = frame;
							byte verticesInside;
							if(AddSurfaceVertices(cube, threshold, out verticesInside))
							{
								AddNeighbouringCubesToList(cubesToCalculate, cube, verticesInside, cubesCalculated);
							}
						}
					}
				}
			}
#if PROFILE
			Program.profiler.EndSection("CalculateSurfaceVertices");
#endif
			doneEvent.Set();
		}

		/// <summary>
		/// Convert a point to X, Y and Z cube indices
		/// </summary>
		/// <param name="point">The point to convert</param>
		/// <param name="cubeX">The X cube index</param>
		/// <param name="cubeY">The Y cube index</param>
		/// <param name="cubeZ">The Z cube index</param>
		private void PointToCubeXYZ(Vector3 point, out int cubeX, out int cubeY, out int cubeZ)
		{
			float fractionX = (point.X - minX) / (maxX - minX);
			cubeX = (int)Math.Floor(fractionX * numCubesX);

			float fractionY = (point.Y - minY) / (maxY - minY);
			cubeY = (int)Math.Floor(fractionY * numCubesY);

			float fractionZ = (point.Z - minZ) / (maxZ - minZ);
			cubeZ = (int)Math.Floor(fractionZ * numCubesZ);
		}

		private readonly int[] neighbourOffsets = { -1, 1, -numCubesZ, numCubesZ, -(numCubesY * numCubesZ), numCubesY * numCubesZ };

		/// <summary>
		/// Given a cube containing the isosurface, add the indices of neighbouring cubes which may also contain it to a list
		/// </summary>
		/// <param name="list">The list to add the neighbour indices to</param>
		/// <param name="cube">The index of the cube whose heighbours are to be added</param>
		/// <param name="verticesInside">A bitfield giving which vertices of the cube are inside the isosurface</param>
		/// <param name="alreadyCalculated">Holds the indeices of the cubes which are not to be added to the list</param>
		private void AddNeighbouringCubesToList(List<int> list, int cube, byte verticesInside, int[] alreadyCalculated)
		{
			Byte neighboursToAdd = neighbourTable[verticesInside];

			for(int i = 0; i < neighbourOffsets.Length; ++i)
			{
				int index = cube + neighbourOffsets[i];

				if(index >= 0 && index < numCubes && alreadyCalculated[index] != frame && (neighboursToAdd & (1 << i)) != 0)
				{
					list.Add(index);
				}
			}
		}

		private ushort[] edgeIndexCache = new ushort[(numCubesX + 1) * (numCubesY + 1) * (numCubesZ + 1) * 3];
		private int[] edgeIndexCached = new int[(numCubesX + 1) * (numCubesY + 1) * (numCubesZ + 1) * 3];

		private ushort[] edgeIndices = new ushort[12];

		/// <summary>
		/// Calculate the position and normals for any surface vertices in a given cube and add these vertices to surfaceVertices
		/// </summary>
		/// <param name="cube">The index of the cube to use</param>
		/// <param name="threshold">The threshold value at which the isosurface lies</param>
		/// <param name="verticesInside">Output: A bitfield giving which vertices of the cube are inside the isosurface</param>
		/// <returns>Whether or not any vertices were added to surfaceVertices</returns>
		private bool AddSurfaceVertices(int cube, float threshold, out byte verticesInside)
		{
			verticesInside = 0;
			int cubeVertexIndex0 = cubeVertexIndices[cube, 0];

			if(vertexValues[cubeVertexIndex0] > threshold)		verticesInside |= 0x01;
			if(vertexValues[cubeVertexIndex0 + 1] > threshold)	verticesInside |= 0x02;
			if(vertexValues[cubeVertexIndex0 + numVerticesZ + 1] > threshold)	verticesInside |= 0x04;
			if(vertexValues[cubeVertexIndex0 + numVerticesZ] > threshold)		verticesInside |= 0x08;
			if(vertexValues[cubeVertexIndex0 + numVerticesY * numVerticesZ] > threshold)		verticesInside |= 0x10;
			if(vertexValues[cubeVertexIndex0 + numVerticesY * numVerticesZ + 1] > threshold)	verticesInside |= 0x20;
			if(vertexValues[cubeVertexIndex0 + numVerticesY * numVerticesZ + numVerticesZ + 1] > threshold)	verticesInside |= 0x40;
			if(vertexValues[cubeVertexIndex0 + numVerticesY * numVerticesZ + numVerticesZ] > threshold)		verticesInside |= 0x80;

			if(verticesInside == 0 || verticesInside == 0xff || surfaceVertexIndex > maxSurfaceVertices - 12)
			{
				return false;
			}

			int cubeX = cube / (numCubesY * numCubesZ);
			int cubeY = (cube / numCubesZ) % numCubesY;
			int cubeZ = cube % numCubesZ;
			int edgeIndexCacheIndex = ((cubeX * (numCubesY + 1) + cubeY) * (numCubesZ + 1) + cubeZ) * 3;

			int usedEdges = edgeTable[verticesInside];

			for(int edge = 0; edge < 12; ++edge)
			{
				if((usedEdges & (1 << edge)) != 0)
				{
					int cacheIndex = edgeIndexCacheIndex + edgeIndexCacheOffsetTable[edge];

					if(edgeIndexCached[cacheIndex] == frame)
					{
						edgeIndices[edge] = edgeIndexCache[cacheIndex];
					}
					else
					{
						int edgeStartVertexIndex = cubeVertexIndices[cube, edgeStartVertexTable[edge]];
						int edgeEndVertexIndex = cubeVertexIndices[cube, edgeEndVertexTable[edge]];

						float t = (threshold - vertexValues[edgeStartVertexIndex]) /
								  (vertexValues[edgeEndVertexIndex] - vertexValues[edgeStartVertexIndex]);
						float oneMinusT = 1.0f - t;

						surfaceVertices[surfaceVertexIndex].position.X =
							oneMinusT * vertexXs[edgeStartVertexIndex] + t * vertexXs[edgeEndVertexIndex];
						surfaceVertices[surfaceVertexIndex].position.Y =
							oneMinusT * vertexYs[edgeStartVertexIndex] + t * vertexYs[edgeEndVertexIndex];
						surfaceVertices[surfaceVertexIndex].position.Z =
							oneMinusT * vertexZs[edgeStartVertexIndex] + t * vertexZs[edgeEndVertexIndex];

						surfaceVertices[surfaceVertexIndex].normal.X =
							oneMinusT * vertexNormalXs[edgeStartVertexIndex] + t * vertexNormalXs[edgeEndVertexIndex];
						surfaceVertices[surfaceVertexIndex].normal.Y =
							oneMinusT * vertexNormalYs[edgeStartVertexIndex] + t * vertexNormalYs[edgeEndVertexIndex];
						surfaceVertices[surfaceVertexIndex].normal.Z =
							oneMinusT * vertexNormalZs[edgeStartVertexIndex] + t * vertexNormalZs[edgeEndVertexIndex];

						edgeIndices[edge] = surfaceVertexIndex;

						edgeIndexCache[cacheIndex] = surfaceVertexIndex;
						edgeIndexCached[cacheIndex] = frame;

						surfaceVertexIndex++;
					}
				}
			}

			for(int triangle = 0; triangleTable[verticesInside, triangle] != -1; triangle += 3)
			{
				if(surfaceIndexIndex <= maxSurfaceIndices - 3)
				{
					surfaceIndices[surfaceIndexIndex++] = edgeIndices[triangleTable[verticesInside, triangle]];
					surfaceIndices[surfaceIndexIndex++] = edgeIndices[triangleTable[verticesInside, triangle + 1]];
					surfaceIndices[surfaceIndexIndex++] = edgeIndices[triangleTable[verticesInside, triangle + 2]];
				}
			}

			return true;
		}

		/// <summary>
		/// Render the isosurface
		/// </summary>
		/// <param name="device">The Direct3D device to render with</param>
		public void RenderSurface(Device device)
		{
			if(surfaceVertexBuffer == null)
			{
				surfaceVertexBuffer = new VertexBuffer(	device, maxSurfaceVertices * Marshal.SizeOf(typeof(SurfaceVertex)),
														Usage.Dynamic | Usage.WriteOnly, SurfaceVertex.Format, Pool.Default);
			}

			if(surfaceIndexBuffer == null)
			{
				surfaceIndexBuffer = new IndexBuffer(	device, maxSurfaceIndices * sizeof(ushort),
														Usage.Dynamic | Usage.WriteOnly, Pool.Default, true);
			}

			if(surfaceVertexIndex > 0 && surfaceIndexIndex > 0)
			{
				DataStream surfaceVertexStream = surfaceVertexBuffer.Lock(	0, surfaceVertexIndex * Marshal.SizeOf(typeof(SurfaceVertex)),
																			LockFlags.Discard);
				surfaceVertexStream.WriteRange(surfaceVertices, 0, surfaceVertexIndex);
				surfaceVertexBuffer.Unlock();

				DataStream surfaceIndexStream = surfaceIndexBuffer.Lock(0, surfaceIndexIndex * sizeof(ushort), LockFlags.Discard);
				surfaceIndexStream.WriteRange(surfaceIndices, 0, surfaceIndexIndex);
				surfaceIndexBuffer.Unlock();

				device.SetStreamSource(0, surfaceVertexBuffer, 0, Marshal.SizeOf(typeof(SurfaceVertex)));
				device.SetIndices(surfaceIndexBuffer);
				device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, surfaceVertexIndex, 0, surfaceIndexIndex / 3);
			}
		}

		/// <summary>
		/// Release any Direct3D default pool resources
		/// </summary>
		public void ReleaseDefaultPoolResources()
		{
			if(surfaceVertexBuffer != null)
			{
				surfaceVertexBuffer.Dispose();
			}
			surfaceVertexBuffer = null;

			if(surfaceIndexBuffer != null)
			{
				surfaceIndexBuffer.Dispose();
			}
			surfaceIndexBuffer = null;
		}

		[StructLayout(LayoutKind.Sequential)]
		private struct GridVertex
		{
			public Vector3 position;

			public static VertexFormat Format
			{
				get
				{
					return VertexFormat.Position;
				}
			}
		}

		/// <summary>
		/// Render the vertices of the grid as a point list
		/// </summary>
		/// <param name="device">The Direct3D device to render with</param>
		public void RenderGrid(Device device)
		{
			GridVertex[] vertices = new GridVertex[numVertices];

			for(int i = 0; i < numVertices; ++i)
			{
				vertices[i].position = new Vector3(vertexXs[i], vertexYs[i], vertexZs[i]);
			}
			
			device.VertexFormat = GridVertex.Format;
			device.DrawUserPrimitives(PrimitiveType.PointList, 0, numVertices, vertices);
		}

		#region Lookup Tables

		static readonly int[] edgeStartVertexTable =
		{
			0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3
		};

		static readonly int[] edgeEndVertexTable =
		{
			1, 2, 3, 0, 5, 6, 7, 4, 4, 5, 6, 7
		};

		static readonly int[] edgeIndexCacheOffsetTable =
		{
			0,
			1 * 3 + 1,
			(numCubesZ + 1) * 3,
			1,
			(numCubesY + 1) * (numCubesZ + 1) * 3,
			((numCubesY + 1) * (numCubesZ + 1) + 1) * 3 + 1,
			((numCubesY + 1) + 1) * (numCubesZ + 1) * 3,
			(numCubesY + 1) * (numCubesZ + 1) * 3 + 1, 
			2,
			1 * 3 + 2,
			((numCubesZ + 1) + 1) * 3 + 2,
			(numCubesZ + 1) * 3 + 2
		};

		static readonly int[] edgeTable =
		{
			0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
			0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
			0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
			0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
			0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
			0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
			0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
			0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
			0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
			0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
			0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
			0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
			0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
			0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
			0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
			0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
			0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
			0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
			0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
			0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
			0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
			0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
			0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
			0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
			0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
			0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
			0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
			0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
			0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
			0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
			0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
			0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0
		};

		static readonly byte[] neighbourTable =
		{
			0x00, 0x15, 0x16, 0x17, 0x1a, 0x1f, 0x1e, 0x1f,
			0x19, 0x1d, 0x1f, 0x1f, 0x1b, 0x1f, 0x1f, 0x0f,
			0x25, 0x35, 0x37, 0x37, 0x3f, 0x3f, 0x3f, 0x3f,
			0x3d, 0x3d, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x2f,
			0x26, 0x37, 0x36, 0x37, 0x3e, 0x3f, 0x3e, 0x3f,
			0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x2f,
			0x27, 0x37, 0x37, 0x33, 0x3f, 0x3f, 0x3f, 0x3b,
			0x3f, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x3f, 0x2b,
			0x2a, 0x3f, 0x3e, 0x3f, 0x3a, 0x3f, 0x3e, 0x3f,
			0x3b, 0x3f, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x2f,
			0x2f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
			0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x2f,
			0x2e, 0x3f, 0x3e, 0x3f, 0x3e, 0x3f, 0x3c, 0x3d,
			0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3d, 0x2d,
			0x2f, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x3d, 0x39,
			0x3f, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x3d, 0x29,
			0x29, 0x3d, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x3f,
			0x39, 0x3d, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x2f,
			0x2d, 0x3d, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
			0x3d, 0x3c, 0x3f, 0x3e, 0x3f, 0x3e, 0x3f, 0x2e,
			0x2f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
			0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x2f,
			0x2f, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x3f, 0x3b,
			0x3f, 0x3e, 0x3f, 0x3a, 0x3f, 0x3e, 0x3f, 0x2a,
			0x2b, 0x3f, 0x3f, 0x3f, 0x3b, 0x3f, 0x3f, 0x3f,
			0x3b, 0x3f, 0x3f, 0x3f, 0x33, 0x37, 0x37, 0x27,
			0x2f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f,
			0x3f, 0x3e, 0x3f, 0x3e, 0x37, 0x36, 0x37, 0x26,
			0x2f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3f, 0x3d, 0x3d,
			0x3f, 0x3f, 0x3f, 0x3f, 0x37, 0x37, 0x35, 0x25,
			0x0f, 0x1f, 0x1f, 0x1b, 0x1f, 0x1f, 0x1d, 0x19,
			0x1f, 0x1e, 0x1f, 0x1a, 0x17, 0x16, 0x15, 0x00
		};

		static readonly int[,] triangleTable =
		{
			{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
			{3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
			{3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
			{3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
			{9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
			{9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
			{2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
			{8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
			{9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
			{4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
			{3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
			{1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
			{4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
			{4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
			{9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
			{5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
			{2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
			{9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
			{0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
			{2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
			{10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
			{4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
			{5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
			{5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
			{9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
			{0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
			{1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
			{10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
			{8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
			{2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
			{7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
			{9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
			{2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
			{11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
			{9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
			{5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
			{11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
			{11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
			{1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
			{9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
			{5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
			{2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
			{0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
			{5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
			{6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
			{3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
			{6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
			{5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
			{1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
			{10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
			{6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
			{8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
			{7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
			{3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
			{5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
			{0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
			{9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
			{8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
			{5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
			{0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
			{6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
			{10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
			{10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
			{8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
			{1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
			{3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
			{0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
			{10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
			{3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
			{6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
			{9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
			{8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
			{3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
			{6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
			{0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
			{10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
			{10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
			{2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
			{7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
			{7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
			{2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
			{1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
			{11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
			{8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
			{0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
			{7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
			{10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
			{2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
			{6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
			{7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
			{2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
			{1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
			{10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
			{10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
			{0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
			{7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
			{6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
			{8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
			{9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
			{6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
			{4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
			{10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
			{8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
			{0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
			{1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
			{8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
			{10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
			{4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
			{10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
			{5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
			{11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
			{9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
			{6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
			{7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
			{3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
			{7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
			{9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
			{3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
			{6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
			{9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
			{1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
			{4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
			{7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
			{6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
			{3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
			{0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
			{6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
			{0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
			{11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
			{6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
			{5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
			{9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
			{1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
			{1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
			{10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
			{0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
			{5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
			{10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
			{11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
			{9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
			{7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
			{2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
			{8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
			{9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
			{9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
			{1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
			{9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
			{9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
			{5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
			{0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
			{10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
			{2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
			{0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
			{0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
			{9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
			{5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
			{3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
			{5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
			{8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
			{0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
			{9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
			{0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
			{1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
			{3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
			{4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
			{9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
			{11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
			{11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
			{2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
			{9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
			{3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
			{1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
			{4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
			{4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
			{0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
			{3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
			{3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
			{0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
			{9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
			{1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
			{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
		};

		#endregion
	}
}